﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace EigenHillCipher
{
    /// <summary>
    /// This is a class to represent SQUARE matrices.
    /// It will represent them and allow you to do a number of actions
    /// upon them and with them.
    /// 
    /// WARNING: Do NOT use this to represent anything other than square
    /// matrices if you want it to work. Additionally, there is no
    /// checking that the matrices are of the same dimensions; it is assumed
    /// by all functions in the class. MAKE SURE you are using two matrices
    /// of the same dimensions under all circumstances.
    /// </summary>
    public class DoubleMatrix
    {
        #region Member Variables

        /// <summary>
        /// This defines the format of numbers being used in the 
        /// matrix when being displayed.
        /// </summary>
        const string NUMBER_FORMAT = "###.00";

        /// <summary>
        /// The rows that comprise this matrix.
        /// </summary>
        private Row[] rows;

        #endregion

        #region Constructors

        /// <summary>
        /// Copy Constructor.
        /// 
        /// Constructs this as a copy of the matrix passed to it.
        /// </summary>
        /// <param name="old">The matrix to clone.</param>
        public DoubleMatrix(DoubleMatrix old)
        {
            rows = new Row[old.rows.Length];
            for (int i=0; i < rows.Length; i++)
            {
                rows[i] = new Row(old.rows[i]);
            }
        }

        /// <summary>
        /// Create a square matrix with the given dimensions. 
        /// Defaults to the multiplicative identity matrix.
        /// </summary>
        /// <param name="size">The number of rows and columns in the matrix.</param>
        public DoubleMatrix(int size)
        {
            rows = new Row[size];
            for (int y = 0; y < size; y++)
            {
                rows[y] = new Row(size);
                for (int x = 0; x < size; x++)
                {
                    if (x == y)
                        this[y, x] = 1.0;
                    else
                        this[y, x] = 0.0;
                }
            }
        }

        /// <summary>
        /// Creates a matrix with the given elements within.
        /// </summary>
        /// <param name="size">The number of rows and columns in the matrix.</param>
        /// <param name="elems">The elements in the matrix. (MUST be the correct number.)</param>
        public DoubleMatrix(int size, params double[] elems)
        {
            rows = new Row[size];
            for (int y = 0; y < size; y++)
            {
                rows[y] = new Row(size);
                for (int x = 0; x < size; x++)
                {
                    rows[y][x] = elems[(size * y)+x];
                }
            }
        }

        #endregion

        #region Special DoubleMatrix Constructors

        /// <summary>
        /// Get the additive identity matrix of the given size.
        /// </summary>
        /// <param name="size">The number of rows and columns in the matrix.</param>
        /// <returns>The additive identity matrix of the given dimensions.</returns>
        public static DoubleMatrix getAdditiveIdentity(int size)
        {
            DoubleMatrix ret = new DoubleMatrix(size);
            for (int i = 0; i < size; i++)
                ret[i, i] = 0;
            return ret;
        }

        /// <summary>
        /// Gets the multiplicative identity matrix of the given size.
        /// </summary>
        /// <param name="size">The number of rows and columns in the matrix.</param>
        /// <returns>The multiplicitive identity matrix of the given dimensions.</returns>
        public static DoubleMatrix getMultiplicativeIdentity(int size)
        {
            return new DoubleMatrix(size);
        }

        #endregion

        #region DoubleMatrix Transformations 

        /// <summary>
        /// Gets the transpose of this matrix.
        /// </summary>
        /// <returns>The transpose of this matrix.</returns>
        public DoubleMatrix getTranspose()
        {
            DoubleMatrix matr = new DoubleMatrix(rows.Length);
            for (int x = 0; x < rows.Length; x++)
                for (int y = 0; y < rows.Length; y++)
                    matr[x, y] = this[y, x];
            return matr;
        }

        /// <summary>
        /// Returns the inverse of this matrix, granted it exists.
        /// Undefined behavior may result if this is called when an inverse does not exist!
        /// </summary>
        /// <returns>The inverse of this matrix, granted it exists.</returns>
        public DoubleMatrix getInverse()
        {
            DoubleMatrix m = new DoubleMatrix(this);

            if (rows.Length == 1)
                return new DoubleMatrix(1, 1 / Determinant);
            else if (rows.Length == 2)
                return (new DoubleMatrix(2, this[1, 1], -this[0, 1], -this[1, 0], this[0, 0]) / Determinant);

            for (int i = 0; i < rows.Length; i++)
                for (int j = 0; j < rows.Length; j++)
                    m[i, j] = Math.Pow(-1.0, Convert.ToDouble(i + j)) * this.Minor(j, i).Determinant;
            m /= Determinant;
            return m;
        }

        #region Usage-Specific

        /// <summary>
        /// Do a random permutation on this matrix, and return the result.
        /// DIAGONAL MATRICES ONLY
        /// </summary>
        /// <param name="rand">The random to use for the permutation.</param>
        /// <returns>This matrix permuted once randomly.</returns>
        public DoubleMatrix permute(ref Random rand)
        {
            DoubleMatrix matr = new DoubleMatrix(this);
            int rnum;
            for (int i = 0; i < Dimension; i++)
            {
                rnum = rand.Next(0, Dimension);
                // Well this would be pretty pointless
                if (rnum != i)
                {
                    double valat = matr[i, i];
                    matr[i, i] = matr[rnum, rnum];
                    matr[rnum, rnum] = valat;
                }
            }

            return matr;
        }

        /// <summary>
        /// Raises this matrix to the given power, optimized diagonal matrices.
        /// 
        /// DIAGONAL MATRICES ONLY
        /// </summary>
        /// <param name="power">The power to raise the matrix to.</param>
        /// <returns>This matrix raised to the power power.</returns>
        public DoubleMatrix raiseToPower(int power)
        {
            /* 
            // Failsafe method
            
            DoubleMatrix matr = new DoubleMatrix(this);

            for (int i = 1; i < power; i++)
                matr *= this;

            return matr;
            */
            // Optimized method
            DoubleMatrix matr = new DoubleMatrix(this);
            
            // Look ma, no braces!
            for (int i = 0; i < Dimension; i++)
                for (int k = 1; k < power; k++)
                    matr[i, i] *= this[i, i];

            return matr;
        }

        /// <summary>
        /// Return the result of multiplying this by a vector.
        /// </summary>
        /// <param name="vector">The vector to multiply.</param>
        /// <param name="N">The value to wrap around.</param>
        /// <returns>A vector of the result.</returns>
        public double[] timesVector(int[] vector)
        {
            double[] result = new double[vector.Length];
            for (int i = 0; i < vector.Length; i++)
            {
                result[i] = 0;
                for (int j = 0; j < vector.Length; j++)
                    result[i] += this[i, j] * vector[j];
            }
            return result;
        }

        /// <summary>
        /// Return the result of multiplying this by a vector.
        /// </summary>
        /// <param name="vector">The vector to multiply.</param>
        /// <param name="N">The value to wrap around.</param>
        /// <returns>A vector of the result.</returns>
        public double[] timesVector(double[] vector)
        {
            double[] result = new double[vector.Length];
            for (int i = 0; i < vector.Length; i++)
            {
                result[i] = 0;
                for (int j = 0; j < vector.Length; j++)
                    result[i] += this[i, j] * vector[j];
            }
            return result;
        }

        #endregion

        #endregion

        #region Public Parameters

        /// <summary>
        /// The determinant of this matrix.
        /// </summary>
        public double Determinant 
        {
            get
            {
                return getDeterminant(this.rows.Length);
            }
        }

        #endregion

        #region Accessors

        /// <summary>
        /// A way to access any element in the matrix.
        /// </summary>
        /// <param name="x">The row that the element to be accessed is in.</param>
        /// <param name="y">The column that the element to be accessed is in.</param>
        /// <returns>The element in the matrix at row x, column y.</returns>
        public double this[int x, int y]
        {
            get { return rows[x][y]; }
            set { rows[x][y] = value; }
        }

        /// <summary>
        /// Gets the dimensions of this matrix.
        /// </summary>
        public int Dimension { get { return rows.Length; } }

        #endregion

        #region Operator and Function Overrides

        /// <summary>
        /// Overrides tostring.
        /// </summary>
        /// <returns>A string representation of this matrix.</returns>
        public override string ToString()
        {
            String str = "";
            for (int i = 0; i < rows.Length; i++)
                str += "[ " + rows[i] + " ]\r\n";
            return str;
        }

        /// <summary>
        /// Overrides tostring, removing the brackets usually shown.
        /// </summary>
        /// <param name="hidebrackets">Whether to remove the brackets or not.</param>
        /// <returns>If showing brackets, same as ToString, otherwise a return-separated list of row vectors in the matrix.</returns>
        public string ToString(bool hidebrackets)
        {
            if (!hidebrackets) return ToString();
            String str = "";
            for (int i = 0; i < rows.Length; i++)
                str += rows[i] + "\r\n";
            return str;
        }


        /// <summary>
        /// Override the times symbol for doubles and matrices.
        /// </summary>
        /// <param name="value">A scalar to multiply by.</param>
        /// <param name="c1">The matrix.</param>
        /// <returns>The result of multiplying the two variables.</returns>
        public static DoubleMatrix operator *(double value, DoubleMatrix c1)
        {
            DoubleMatrix c3 = new DoubleMatrix(c1);
            for (int i = 0; i < c3.rows.Length; i++)
                c3.rows[i] *= value;
            return c3;
        }
        /// <summary>
        /// Override the times symbol for doubles and matrices.
        /// </summary>
        /// <param name="c1">The matrix.</param>
        /// <param name="value">The scalar to multiply by.</param>
        /// <returns>The result of the multiplication.</returns>
        public static DoubleMatrix operator *(DoubleMatrix c1, double value)
        {
            return value * c1;
        }

        /// <summary>
        /// Override the times symbol for multiplying two matrices.
        /// </summary>
        /// <param name="c1">The first matrix to multiply. (Left)</param>
        /// <param name="c2">The second matrix to multiply. (Right)</param>
        /// <returns>The result of the multiplication.</returns>
        public static DoubleMatrix operator *(DoubleMatrix c1, DoubleMatrix c2)
        {
            DoubleMatrix c3 = DoubleMatrix.getAdditiveIdentity(c1.rows.Length);
            for (int x = 0; x < c1.rows.Length; x++)
            {
                for (int y = 0; y < c1.rows.Length; y++)
                {
                    // c3[x, y] = 0;
                    for (int i = 0; i < c1.rows.Length; i++)
                        c3[x, y] += c1[x, i] * c2[i, y];
                }
            }
            return c3;
        }

        /// <summary>
        /// Division of a matrix by a scalar.
        /// </summary>
        /// <param name="c1">The matrix.</param>
        /// <param name="value">The scalar.</param>
        /// <returns>The result of dividing c1 by value.</returns>
        public static DoubleMatrix operator /(DoubleMatrix c1, double value)
        {
            DoubleMatrix c3 = new DoubleMatrix(c1);
            for (int i = 0; i < c3.rows.Length; i++)
                c3.rows[i] /= value;
            return c3;
        }

        #endregion

        #region Private Helper Functions

        /// <summary>
        /// Get the determinant of this matrix. (Recursive piece.)
        /// </summary>
        /// <param name="n">The dimension of the matrix with which we are concerned.</param>
        /// <returns></returns>
        private double getDeterminant(int n)
        {
            double det = 0;
            DoubleMatrix nu = new DoubleMatrix(n);
            int m = 0;
            if (rows.Length == 1)
                return this[0, 0];
            else if (rows.Length == 2)
                return this[0, 0] * this[1, 1] - this[0, 1] * this[1, 0];
            else
            {
                for (int k = 0; k < n; k++)
                {
                    for (int i = 1; i < n; i++)
                    {
                        m = 0;
                        for (int j = 0; j < n; j++)
                        {
                            if (j == k)
                                continue;
                            nu[i - 1, m] = this[i, j];
                            m++;
                        }
                    }
                    det += Math.Pow(-1.0, k + 2.0) * this[0, k] * nu.getDeterminant(n-1);
                }
                return det;
            }
        }

        /// <summary>
        /// Grab the minor of this matrix at position x,y.
        /// </summary>
        /// <param name="x">The row of the matrix.</param>
        /// <param name="y">The column of the matrix.</param>
        /// <returns>The minor requested.</returns>
        private DoubleMatrix Minor(int x, int y)
        {
            DoubleMatrix nu = new DoubleMatrix(rows.Length - 1);
            int xx = 0, yy = 0;
            for (int i=0; i < rows.Length; i++)
                if (i != x)
                {
                    for (int j=0; j < rows.Length; j++)
                        if (j != y)
                        {
                            nu[xx, yy] = this[i, j];
                            yy++;
                        }
                    yy = 0;
                    xx++;
                }
            return nu;   
        }

        /// <summary>
        /// Shorcut to swapping rows in the matrix.
        /// </summary>
        /// <param name="row1">The index of the first row to swap.</param>
        /// <param name="row2">The index of the second row to swap.</param>
        private void swap(int row1, int row2)
        {
            Row temp = rows[row1];
            rows[row1] = rows[row2];
            rows[row2] = temp;
        }

        /// <summary>
        /// Grab a double-array version of a column.
        /// </summary>
        /// <param name="num">Index of the column.</param>
        /// <returns>Column #num.</returns>
        private double[] getColumn(int num)
        {
            double[] ret = new double[rows.Length];
            for (int i = 0; i < rows.Length; i++)
                ret[i] = this[i, num];
            return ret;
        }

        /// <summary>
        /// Grab a double-array version of a row.
        /// </summary>
        /// <param name="num">The index of the row.</param>
        /// <returns>Row #num</returns>
        private double[] getRow(int num)
        {
            return rows[num].getArray();
        }

        /// <summary>
        /// Get the dot product of two vectors stored as doubles.
        /// </summary>
        /// <param name="a">The first vector.</param>
        /// <param name="b">The second vector.</param>
        /// <returns>The dot product of a and b as vectors.</returns>
        private static double Dot(double[] a, double[] b)
        {
            double ret = 0;
            for (int i = 0; i < a.Length; i++)
                ret += a[i] * b[i];
            return ret;
        }

        #endregion

        #region Row Class; only used internally.

        /// <summary>
        /// A class used to represent rows in the matrix.
        /// 
        /// Created to ease all of the row operations involved with
        /// the functions being implemented.
        /// </summary>
        private class Row
        {

            #region Private Members

            /// <summary>
            /// The elements in this row.
            /// </summary>
            double[] elements;

            #endregion

            #region Constructors

            /// <summary>
            /// Construct a copy of an existing row.
            /// </summary>
            /// <param name="old">The row to make this a copy of.</param>
            public Row(Row old)
            {
                elements = new double[old.elements.Length];
                for (int i = 0; i < elements.Length; i++)
                    elements[i] = old.elements[i];
            }
            
            /// <summary>
            /// Create a row with SIZE elements.
            /// </summary>
            /// <param name="size">The number of elements in this row.</param>
            public Row(int size)
            {
                elements = new double[size];
            }

            /// <summary>
            /// Create a row containing the elements provided.
            /// </summary>
            /// <param name="elems">The elements for this row.</param>
            public Row(params double[] elems)
            {
                elements = elems;
            }

            #endregion

            #region Accessors

            /// <summary>
            /// This returns a double array of the elements in this. 
            /// Realistically this just returns the underlying row structure.
            /// </summary>
            /// <returns>The underlying structure of elements.</returns>
            public double[] getArray()
            {
                return elements;
            }

            /// <summary>
            /// Get the element at position i within the row.
            /// </summary>
            /// <param name="i">The position to grab the element at.</param>
            /// <returns>The element at position i in this row.</returns>
            public double this[int i]
            {
                get { return elements[i]; }
                set { elements[i] = value; }
            }

            #endregion

            #region Method and Operator Overrides

            /// <summary>
            /// Override the division operator for Rows.
            /// </summary>
            /// <param name="c1">The row to divide.</param>
            /// <param name="value">The scalar to divide by.</param>
            /// <returns>A row representing the result of the division by the scalar.</returns>
            public static Row operator /(Row c1, double value)
            {
                Row c3 = new Row(c1);
                for (int i = 0; i < c3.elements.Length; i++)
                    c3.elements[i] /= value;
                return c3;

            }

            /// <summary>
            /// Override multiplication of scalars by rows.
            /// </summary>
            /// <param name="c1">The row to multiply.</param>
            /// <param name="value">The scalar to multiply this by.</param>
            /// <returns>The result of the multiplication.</returns>
            public static Row operator *(Row c1, double value)
            {
                Row c3 = new Row(c1);
                for (int i = 0; i < c3.elements.Length; i++)
                    c3.elements[i] *= value;
                return c3;
            }

            /// <summary>
            /// Override the + operator to add two rows together.
            /// </summary>
            /// <param name="c1">The first row.</param>
            /// <param name="c2">The second row.</param>
            /// <returns>The result of summing the elements in the two rows.</returns>
            public static Row operator +(Row c1, Row c2)
            {
                Row c3 = new Row(c1);
                for (int i = 0; i < c1.elements.Length; i++)
                    c3[i] += c2[i];
                return c3;
            }

            /// <summary>
            /// Override the - operator to subtract two rows from eachother.
            /// </summary>
            /// <param name="c1">The row to subtract from.</param>
            /// <param name="c2">The row to subtract.</param>
            /// <returns>The result of c1-c2.</returns>
            public static Row operator -(Row c1, Row c2)
            {
                Row c3 = new Row(c1);
                for (int i = 0; i < c1.elements.Length; i++)
                {
                    c3[i] -= c2[i];
                }
                return c3;
            }

            /// <summary>
            /// Override the ToString method.
            /// </summary>
            /// <returns>Returns a human-readable string format of this row.</returns>
            public override string ToString()
            {
                String str = "";
                foreach(Double element in elements)
                {
                    str += element.ToString(NUMBER_FORMAT).PadLeft(NUMBER_FORMAT.Length,' ') + " ";
                }
                str = str.Substring(0, str.Length - 1);
                return str;
            }

            #endregion
        }

        #endregion
    }
}
