﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace EigenHillCipher
{
    public partial class MatrixSys : Form
    {
        /// <summary>
        /// The alphabet cardinality; N from the paper.
        /// </summary>
        const int ALPHAMAX = 256;

        /// <summary>
        /// The maximum power to raise matrices to. 
        /// 
        /// This decision was made to decrease computation time
        /// and reduce the liklihood of precision errors with the
        /// doubles usedto store the values.
        /// </summary>
        const int POWERMAX = 25;

        /// <summary>
        /// A value to compare doubles with to find out if it is
        /// close enough to zero to be considered singular in terms
        /// of our application.
        /// </summary>
        const double EPSILON = .001;

        /// <summary>
        /// Is this currently being dragged?
        /// </summary>
        bool dragging = false;

        /// <summary>
        /// This is the position on the control that was clicked
        /// by the user. This is used to make dragging the window
        /// work right.
        /// </summary>
        int start_mousex, start_mousey;

        /// <summary>
        /// Global variables Q Q^-1 and D
        /// </summary>
        DoubleMatrix Q, Qprime, D;

        /// <summary>
        /// The eigenvalues being used.
        /// </summary>
        int[] eigenvalues;

        /// <summary>
        /// The doubles being ciphered.
        /// </summary>
        double[] cipherdoubles;

        /// <summary>
        /// The random number seeds used.
        /// </summary>
        int seedl, seedt;

        /// <summary>
        /// The plaintext used to cipher, and the resultant decoded
        /// text.
        /// </summary>
        String plaintext, decodedciphertext;
        
        public MatrixSys()
        {
            InitializeComponent();

            // Set our encoding and decoding matrices.
            Q = new DoubleMatrix(3, 1, -1, 3, 2, 1, 2, -2, -2, 1);
            Qprime = Q.getInverse();

            // Set our two random seeds... RANDOMLY!
            Random siderand = new Random();
            seedl = siderand.Next();
            seedt = siderand.Next();

            // Generate random applicable eigenvalues.
            eigenvalues = generateEigens(Q.Dimension);

            // Default string to encode.
            plaintext = "This is a modified hill cipher algorithm.";
            decodedciphertext = "Not decoded yet.";

            refreshSystem();
        }

        // ugly way of generating eigenvalues
        private int[] generateEigens(int num)
        {
            // Create a new array...
            int[] ret = new int[num];
            int found = 0;
            Random rand = new Random();


            while (found < num)
            {
                // Fill it with SEQUENCIAL eigenvalues
                found = 0;
                for (int i = rand.Next(1, ALPHAMAX); i < ALPHAMAX; i++)
                    if (gcd(i, ALPHAMAX) == 1 && !ret.Contains<int>(i))
                    {
                        ret[found++] = i;
                        if (found >= num) break;
                    }
            }
            return ret;
        }

        /// <summary>
        /// Refreshes everything in the system.
        /// </summary>
        private void refreshSystem()
        {
            // Make sure the matrix can be inverted.
            if (Math.Abs(Q.Determinant) < EPSILON)
            {
                Inputform.showMessageDialog("Error","The matrix provided is not invertible, or appears singular to machine precision. Please pick another matrix. Encryption aborted.");
            }
            else
            {
                doEncryption();
                doDecryption();
                Refresh();
            }
        }

        /// <summary>
        /// Close button
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button1_Click(object sender, EventArgs e)
        {
            Close();
        }

        
        private void MatrixSys_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Escape)
                this.Close();
        }

        /// <summary>
        /// Start dragging
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void moveBar_MouseDown(object sender, MouseEventArgs e)
        {
            dragging = true;
            start_mousex = e.X;
            start_mousey = e.Y;
        }

        /// <summary>
        /// Stop dragging
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void moveBar_MouseUp(object sender, MouseEventArgs e)
        {
            dragging = false;
        }

        /// <summary>
        /// Move the form if we are dragging.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MatrixSys_MouseMove(object sender, MouseEventArgs e)
        {
            if (dragging)
            {
                this.Top = MousePosition.Y - start_mousey;
                this.Left = MousePosition.X - start_mousex;
                Invalidate();
            }
        }

        /// <summary>
        /// Show the information dialog.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void InfoButton_Click(object sender, EventArgs e)
        {
            (new AboutBox()).ShowDialog(this);
        }

        /// <summary>
        /// Update everything that is shown on the form.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MatrixSys_Paint(object sender, PaintEventArgs e)
        {
            encodingMatrix.Text = Q.ToString(true);
            decodingMatrix.Text = Qprime.ToString(true);
            lblCiphertext.Text = outDoubleArray(cipherdoubles);
            lblDecodedText.Text = "".PadLeft(6) + decodedciphertext.Replace('\0','&');
            lblPlainText.Text = "".PadLeft(6) + plaintext;
            lblSeedL.Text = seedl.ToString();
            lblSeedT.Text = seedt.ToString();
            lblEigens.Text = outIntArray(eigenvalues);
            lblDiag.Text = D.ToString(true);
        }

        /// <summary>
        /// Function to get the greatest common divisor of two numbers.
        /// </summary>
        /// <param name="a">First number.</param>
        /// <param name="b">Second number.</param>
        /// <returns>The GCD of these two numbers.</returns>
        private int gcd(int a, int b)
        {
            int temp;
            while (b != 0)
            {
                temp = b;
                b = a % b;
                a = temp;
            }
            return a;
        }

        /// <summary>
        /// Gets the euler's totient number for the number given.
        /// NOTE: This is about as efficient as carrying a bucket of water upside down.
        /// Use with caution.
        /// 
        /// THIS IS UNUSED.
        /// </summary>
        /// <param name="n">The number given.</param>
        /// <returns>The result of euler's totient of it.</returns>
        private int eulerTotient(int n)
        {
            int sum = 0;
            for (int i = 1; i <= n; i++)
                if (gcd(i, n) == 1) 
                    sum++;
            return sum;
        }

        /// <summary>
        /// Do the actual encryption on the system.
        /// </summary>
        private void doEncryption()
        {
            
            // Construct Diagonal Matrix
            D = new DoubleMatrix(Q.Dimension);
            for (int i = 0; i < D.Dimension; i++)
                D[i, i] = eigenvalues[i];

            // Seed two random number generators...
            // Random numbers.
            Random PRNGL = new Random(seedl);
            
            // Random permutations
            Random PRPGT = new Random(seedt);

            // Key matrix (There is NOT a single key matrix. 
            // The paper is misleading in this.
            // DoubleMatrix K = Q * D * Qprime;

            // Permuted version of the matrix.
            DoubleMatrix Dperm = D;
            
            // Force the ciphertext to be of a length evenly divisible by the block size.
            plaintext = plaintext.PadRight(plaintext.Length+(plaintext.Length % Q.Dimension));

            // The ciphered numbers.
            cipherdoubles = new double[plaintext.Length];

            // li from the paper. Set to random numbers from seedl.
            int li; 

            for (int i = 0; i < plaintext.Length / Q.Dimension; i++)
            {
                li = PRNGL.Next(1, POWERMAX);

                // Collect the block of text and put it into ints.
                int[] block = new int[Q.Dimension];
                for (int j = 0; j < Q.Dimension; j++)
                    block[j] = plaintext[(i * Q.Dimension + j)] % ALPHAMAX;

                // Randomly permute the diagonal matrix.
                Dperm = Dperm.permute(ref PRPGT);

                // Calculate the key matrix at this position.
                DoubleMatrix KMi = Q * ( Dperm.raiseToPower(li) ) * Qprime;
                
                // Encipher the current block by multiplying with KMi
                double[] cipherblock = KMi.timesVector(block);

                // Add eigenvalues from d in order.
                for (int j = 0; j < D.Dimension; j++)
                    cipherblock[j] += Dperm[j, j];

                // Put the values into the collection of doubles
                for (int j = 0; j < Q.Dimension; j++)
                    cipherdoubles[i*Q.Dimension+j] = cipherblock[j];

                // Rinse, Lather, Repeat.
            }

        }

        /// <summary>
        /// Decrypt the thing we just encrypted.
        /// </summary>
        private void doDecryption()
        {

            // Construct Diagonal Matrix
            D = new DoubleMatrix(Q.Dimension);
            for (int i = 0; i < D.Dimension; i++)
                D[i, i] = eigenvalues[i];

            // Construct Dprime (Just in case.)
            DoubleMatrix Dprime = D.getInverse();

            // Seed the pseudorandom number generator
            Random PRNGL = new Random(seedl);

            // For decoding, we need to permute both the original diagonalization matrix,
            // AND the inverted one. The easiest way I see to do this is to diagonalize both
            // using the same seed, using the same logic this is based on.
            Random PRPGT = new Random(seedt);
            Random PRPGTB = new Random(seedt);

            // This is li from the equation.
            int li;

            // Set up a copy of D to permute, and a copy of 
            // dprime as well. (Both are used.)
            DoubleMatrix Dperm = D, Dprimeperm = Dprime;

            // Clear the decoded cipher text.
            decodedciphertext = "";


            for (int i = 0; i < cipherdoubles.Length / Q.Dimension; i++)
            {
                // Set li to a random number < powermax
                li = PRNGL.Next(1, POWERMAX);

                // Get the block we are decoding out.
                double[] block = new double[Q.Dimension];
                for (int j = 0; j < Q.Dimension; j++)
                    block[j] = cipherdoubles[(i * Q.Dimension + j)];

                // Permute D and Dprime.
                Dperm = Dperm.permute(ref PRPGT);
                Dprimeperm = Dprimeperm.permute(ref PRPGTB);

                // Subtract eigenvalues from d in order.
                for (int j = 0; j < Q.Dimension; j++)
                    block[j] -= Dperm[j, j];

                // Calculate the current key matrix.
                DoubleMatrix KMiprime = Q * (Dprimeperm.raiseToPower(li)) * Qprime;

                // Store the deciphered text in this.
                double[] decipherblock = KMiprime.timesVector(block);

                // Put the decoded text into the decoded string.
                for (int j = 0; j < decipherblock.Length; j++)
                    decodedciphertext += Convert.ToChar(Convert.ToInt32(Math.Round(Math.Abs(decipherblock[j] % ALPHAMAX))));

            }
        }

        /// <summary>
        /// Output a double array in string format. 
        /// </summary>
        /// <param name="array">The array to print out.</param>
        /// <returns>A formatted string of the double.</returns>
        private string outDoubleArray(double[] array)
        {
            if (array == null) return "";
            string ret = "";
            for (int i = 0; i < array.Length; i++)
                ret += array[i].ToString() + ", ";
            ret = ret.Substring(0, ret.Length - 2);
            // ret += "]";
            return ret;
        }

        /// <summary>
        /// Output an int array to a string.
        /// </summary>
        /// <param name="array">The array to print out.</param>
        /// <param name="hidebrackets">Whether to hide the brackets around it.</param>
        /// <returns>A formatted string of the ints.</returns>
        private string outIntArray(int[] array, bool hidebrackets = false)
        {
            string ret = "";
            if (array == null) return ret;
            if (!hidebrackets)
                ret = "[";

            for (int i = 0; i < array.Length; i++)
                ret += array[i].ToString() + ", ";
            ret = ret.Substring(0, ret.Length - 2);
            if (!hidebrackets)
                ret += "]";
            return ret;
        }

        /// <summary>
        /// change seedl
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lblSeedL_Click(object sender, EventArgs e)
        {
            string strseedl = seedl.ToString();
            int oldseedl = seedl;
            Inputform.showInputDialog("Set SeedL", "Please enter a value for the seed for the random number generator.", ref strseedl);
            try
            {
                seedl = Convert.ToInt32(strseedl);
            }
            catch (FormatException)
            {
                seedl = oldseedl;
            }
            catch (OverflowException)
            {
                seedl = oldseedl;
            }
            refreshSystem();
        }

        /// <summary>
        /// Change seedt.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lblSeedT_Click(object sender, EventArgs e)
        {
            string strseedt = seedt.ToString();
            int oldseedt = seedt;
            Inputform.showInputDialog("Set SeedL", "Please enter a value for the seed for the random permutation generator, used for permuting the diagonalization matrix.", ref strseedt);
            try
            {
                seedt = Convert.ToInt32(strseedt);
            }
            catch (FormatException)
            {
                seedt = oldseedt;
            }
            catch (OverflowException)
            {
                seedt = oldseedt;
            }
            refreshSystem();
        }

        /// <summary>
        /// Change the text to encrypt
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lblPlainText_Click(object sender, EventArgs e)
        {
            string newmessage = plaintext;
            Inputform.showInputDialog("Set Text", "Please enter the text you would like to have encrypted below.", ref newmessage);
            if (newmessage.Length > 0)
                plaintext = newmessage;
            refreshSystem();
        }

        /// <summary>
        /// Allow the user to input the text to encode.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void encodingMatrix_Click(object sender, EventArgs e)
        {
            string oldmatrix = tomatlab(Q);
            Inputform.showInputDialog("Enter Q", "Please enter the invertible matrix you would like to use to encode the text. Use MATLAB format.", ref oldmatrix);
            
            // The lines below this point are rather unusual, but functional.
            if (!frommatlab(oldmatrix, ref Q))
                while (
                    Inputform.showInputDialog("Enter Encoding Matrix", "Your matrix was not square, or not properly formatted. Please fix it, or hit cancel to continue using the old matrix.", ref oldmatrix)
                        == DialogResult.OK && !frommatlab(oldmatrix, ref Q)
                      )
                    ; // The while covers everything.
            Qprime = Q.getInverse();
            eigenvalues = generateEigens(Q.Dimension);
            D = new DoubleMatrix(Q.Dimension);
            for (int i = 0; i < Q.Dimension; i++)
                D[i, i] = eigenvalues[i];
            refreshSystem();
        }

        /// <summary>
        /// Converts a matrix to MATLAB-like formatting so the user can edit them.
        /// </summary>
        /// <param name="m">The matrix.</param>
        /// <returns>A string representation of m.</returns>
        private string tomatlab(DoubleMatrix m)
        {
            string format = "";
            for (int i = 0; i < m.Dimension; i++)
            {
                for (int j= 0; j < m.Dimension; j++)
                    format += m[i,j].ToString() + " ";
                format += " ; ";
            }
            if (format.Length > 3)
                format = format.Substring(0,format.Length - 3);
            return format;
        }

        /// <summary>
        /// Attempts to read a matrix from a string in MATLAB format.
        /// </summary>
        /// <param name="text">The text to convert to a matrix.</param>
        /// <param name="mbox">The matrix to copy into, if successful. 
        ///                    This remains untouched if the result is false.</param>
        /// <returns>True if successful, false if an error occurs.</returns>
        private bool frommatlab(string text, ref DoubleMatrix mbox)
        {
            string[] split = text.Split(new char[] { ' ', ';' }, StringSplitOptions.RemoveEmptyEntries);
            double dlen = Math.Sqrt(split.Length);
            int ilen = Convert.ToInt32(Math.Round(dlen));
            
            // Make SURE this is a proper square matrix.
            if (ilen * ilen != split.Length)
                return false;
            DoubleMatrix temp = new DoubleMatrix(ilen);

            try
            {
                for (int i = 0; i < ilen; i++)
                    for (int j = 0; j < ilen; j++)
                        temp[i, j] = Convert.ToInt32(split[i * ilen + j]);
            }
            catch (FormatException)
            {
                return false;
            }
            catch (OverflowException)
            {
                return false;
            }

            mbox = temp;
            return true;
        }

        /// <summary>
        /// Change the eigenvalues
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lblEigens_Click(object sender, EventArgs e)
        {
            string oldeigens = outIntArray(eigenvalues, true);
            Inputform.showInputDialog("Enter Eigenvalues", "Please enter a sequencial comma-separated list of eigenvalues for the matrix. They must be coprime with " + ALPHAMAX + " and greater than 0.", ref oldeigens);

            // The lines below this point are rather unusual, but functional.
            if (!doublesfromlist(oldeigens, Q.Dimension, ref eigenvalues))
                while (
                    Inputform.showInputDialog("Enter Eigenvalues", "Your did not enter the right number of eigenvalues, or one of them was not coprime with " + ALPHAMAX + ". Please try again.", ref oldeigens)
                        == DialogResult.OK && !doublesfromlist(oldeigens, Q.Dimension, ref eigenvalues)
                      )
                    ; // The while covers everything.
            Qprime = Q.getInverse();
            for (int i = 0; i < Q.Dimension; i++)
                D[i, i] = eigenvalues[i];
            refreshSystem();
        }

        /// <summary>
        /// This grabs a collection of doubles from a string.
        /// </summary>
        /// <param name="text">The string to convert to doubles.</param>
        /// <param name="dimm">The required size of the array.</param>
        /// <param name="result">The double array to copy into.</param>
        /// <returns>True if successful, false otherwise.</returns>
        private bool doublesfromlist(string text, int dimm, ref int[] result)
        {
            string[] split = text.Split(new char[] { ' ', ';',',' }, StringSplitOptions.RemoveEmptyEntries);

            // Make SURE this is a proper square matrix.
            if (split.Length != dimm)
                return false;
            int[] temp = new int[dimm];
            try
            {
                for (int i = 0; i < dimm; i++)
                {
                    int tempi = Convert.ToInt32(split[i]);
                    if (tempi <= 0 || gcd(tempi, ALPHAMAX) != 1 || temp.Contains<int>(tempi))
                        throw new FormatException();
                    temp[i] = tempi;
                }
            }
            catch (FormatException)
            {
                return false;
            }
            catch (OverflowException)
            {
                return false;
            }

            result = temp;
            return true;
        }

        // Get doubles from a list [no bounding]
        private bool doublesfromlist(string text, ref double[] result)
        {
            string[] split = text.Split(new char[] { ' ', ';', ',' }, StringSplitOptions.RemoveEmptyEntries);

            double[] temp = new double[split.Length];
            try
            {
                for (int i = 0; i < split.Length; i++)
                {
                    double tempi = Convert.ToDouble(split[i]);
                    temp[i] = tempi;
                }
            }
            catch (FormatException)
            {
                return false;
            }
            catch (OverflowException)
            {
                return false;
            }

            result = temp;
            return true;
        }

        /// <summary>
        /// Copy button
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button2_Click(object sender, EventArgs e)
        {
            Clipboard.SetText(outDoubleArray(cipherdoubles));
        }

        /// <summary>
        /// Decode a premade cipher.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button3_Click(object sender, EventArgs e)
        {
            string matr = tomatlab(Q);
            if (Inputform.showInputDialog("Enter Matrix", "Please enter the matrix to use for decoding.", ref matr) != DialogResult.OK)
                return;

            if (!frommatlab(matr, ref Q))
            {
                Inputform.showMessageDialog("Error", "Error: Could not properly translate matrix! Aborting...");
                return;
            }

            Qprime = Q.getInverse();

            string eigens = outIntArray(eigenvalues, true);
            if (Inputform.showInputDialog("Enter Eigenvalues", "Please enter the eigenvalues used.", ref eigens) != DialogResult.OK)
                return;

            if (!doublesfromlist(eigens, Q.Dimension, ref eigenvalues))
            {
                Inputform.showMessageDialog("Error", "Error: Could not properly translate eigenvalues! Aborting...");
                return;
            }

            string _seedl = seedl.ToString();
            string _seedt = seedt.ToString();

            if ( Inputform.showInputDialog("Enter SeedL","Please enter the value for SeedL", ref _seedl) != DialogResult.OK ||
                 Inputform.showInputDialog("Enter SeedT","Please enter the value for SeedT", ref _seedt) != DialogResult.OK)
                return;

            try
            {
                seedl = Convert.ToInt32(_seedl);
                seedt = Convert.ToInt32(_seedt);
            }
            catch (FormatException)
            {
                Inputform.showMessageDialog("Error", "Error: Could not properly translate seed values! Aborting...");
                return;
            }
            catch (OverflowException)
            {
                Inputform.showMessageDialog("Error", "Error: Could not properly translate seed values! Aborting...");

                return;
            }

            string decoded = "";
            if (Inputform.showInputDialog("Enter Cipher Text", "Please enter the ciphertext as copied from the program.", ref decoded) != DialogResult.OK)
                return;

            if (!doublesfromlist(decoded, ref cipherdoubles))
            {
                Inputform.showMessageDialog("Error", "Could not parse ciphertext. Aborting.");
                return;
            }

            plaintext = "Not set";

            doDecryption();
            Invalidate();



        }
    }
}
